import re

from vladiate.exceptions import ValidationException, BadValidatorException


class Validator(object):
    ''' Generic Validator class '''

    def __init__(self):
        self.fail_count = 0

    @property
    def bad(self):
        ''' Return something containing the "bad" fields '''
        raise NotImplementedError

    def validate(self, field, row):
        ''' Validate the given field. Also is given the row context '''
        raise NotImplementedError


class CastValidator(Validator):
    ''' Validates that a field can be cast to a float '''

    def __init__(self):
        super(CastValidator, self).__init__()
        self.invalid_set = set([])

    def validate(self, field, row={}):
        try:
            if (field or not self.empty_ok):
                self.cast(field)
        except ValueError as e:
            self.invalid_set.add(field)
            raise ValidationException(e)

    @property
    def bad(self):
        return self.invalid_set


class FloatValidator(CastValidator):
    ''' Validates that a field can be cast to a float '''

    def __init__(self, empty_ok=False):
        super(FloatValidator, self).__init__()
        self.empty_ok = empty_ok
        self.cast = float


class IntValidator(CastValidator):
    ''' Validates that a field can be cast to an int '''

    def __init__(self, empty_ok=False):
        super(IntValidator, self).__init__()
        self.empty_ok = empty_ok
        self.cast = int


class SetValidator(Validator):
    ''' Validates that a field is in the given set '''

    def __init__(self, valid_set=[], empty_ok=False):
        super(SetValidator, self).__init__()
        self.valid_set = set(valid_set)
        self.invalid_set = set([])
        if empty_ok:
            self.valid_set.add('')

    def validate(self, field, row={}):
        if field not in self.valid_set:
            self.invalid_set.add(field)
            raise ValidationException(
                "'{}' is not in {}".format(field, self.valid_set))

    @property
    def bad(self):
        return self.invalid_set


class UniqueValidator(Validator):
    ''' Validates that a field is unique within the file '''

    def __init__(self, unique_with=[]):
        super(UniqueValidator, self).__init__()
        self.unique_values = set([])
        self.duplicates = set([])
        self.unique_with = unique_with
        self.unique_check = False

    def _precheck_unique_with(self, row):
        extra = set(self.unique_with) - set(row.keys())
        if extra:
            raise BadValidatorException(extra)
        self.unique_check = True

    def validate(self, field, row={}):
        if self.unique_with and not self.unique_check:
            self._precheck_unique_with(row)

        key = tuple([field] + [row[k] for k in self.unique_with])
        if key not in self.unique_values:
            self.unique_values.add(key)
        else:
            self.duplicates.add(key)
            if self.unique_with:
                raise ValidationException(
                    "'{}' is already in the column (unique with: {})".format(
                        field, key[1:]))
            else:
                raise ValidationException(
                    "'{}' is already in the column".format(field))

    @property
    def bad(self):
        return self.duplicates


class RegexValidator(Validator):
    ''' Validates that a field matches a given regex '''

    def __init__(self, pattern=r'di^', empty_ok=False):
        super(RegexValidator, self).__init__()
        self.regex = re.compile(pattern)
        self.empty_ok = empty_ok
        self.failures = set([])

    def validate(self, field, row={}):
        if not self.regex.match(field) and (field or not self.empty_ok):
            self.failures.add(field)
            raise ValidationException(
                "'{}' does not match pattern /{}/".format(field, self.regex))

    @property
    def bad(self):
        return self.failures


class RangeValidator(Validator):
    def __init__(self, low, high):
        self.fail_count = 0
        self.low = low
        self.high = high
        self.outside = set()

    def validate(self, field, row={}):
        if not self.low <= float(field) <= self.high:
            self.outside.add(field)
            raise ValidationException(
                "'{}' is not in range {} to {}".format(
                    field, self.low, self.high
                )
            )

    @property
    def bad(self):
        return self.outside


class EmptyValidator(Validator):
    ''' Validates that a field is always empty '''

    def __init__(self):
        super(EmptyValidator, self).__init__()
        self.nonempty = set([])

    def validate(self, field, row={}):
        if field != '':
            self.nonempty.add(field)
            raise ValidationException(
                "'{}' is not an empty string".format(field)
            )

    @property
    def bad(self):
        return self.nonempty


class NotEmptyValidator(Validator):
    ''' Validates that a field is not empty '''

    def __init__(self):
        self.fail_count = 0

    def validate(self, field, row={}):
        if field == '':
            raise ValidationException("Row has emtpy field in column")

    @property
    def bad(self):
        # Return an empty set to conform to the protocol. The 'bad' fields
        # would all be empty strings anyways
        return set()


class Ignore(Validator):
    ''' Ignore a given field. Never fails '''

    def validate(self, field, row={}):
        pass

    @property
    def bad(self):
        pass



import pytest
from pretend import stub, call, call_recorder

from ..validators import (
    CastValidator, EmptyValidator, FloatValidator, Ignore, IntValidator,
    NotEmptyValidator, RangeValidator, RegexValidator, SetValidator,
    UniqueValidator, Validator
)
from ..exceptions import BadValidatorException, ValidationException


class FakeRow(object):
    ''' A row with side effects '''

    def __init__(self, fields):
        self.fields = fields

    def __getitem__(self, key):
        return self.fields[key].pop(0)

    def keys(self):
        return self.fields.keys()


def test_cast_validator():

    def _raise_valueerror(x):
        raise ValueError

    validator = CastValidator()
    validator.cast = call_recorder(_raise_valueerror)

    assert validator.bad == set()

    with pytest.raises(ValidationException):
        validator.validate('field', 'something')

    assert validator.bad == {'field'}

    assert validator.cast.calls == [
        call('field')
    ]


@pytest.mark.parametrize('field', [('42'), ('42.0'), ('-42.0'), ])
def test_float_validator_works(field):
    FloatValidator().validate(field)


@pytest.mark.parametrize('field', [('foo'), (' '), ('7..0'), ('4,200'), ])
def test_float_validator_fails(field):
    with pytest.raises(ValidationException):
        FloatValidator().validate(field)


@pytest.mark.parametrize('field', [('42'), ])
def test_int_validator_works(field):
    IntValidator().validate(field)


@pytest.mark.parametrize('field', [
    ('foo'),
    (' '),
    ('7..0'),
    ('4,200'),
    ('42.0'),
    ('-42.0'),
])
def test_int_validator_fails(field):
    with pytest.raises(ValidationException):
        IntValidator().validate(field)


@pytest.mark.parametrize('field_set, field', [
    (['foo'], 'foo'),
    (['foo', 'bar'], 'foo'),
])
def test_set_validator_works(field_set, field):
    validator = SetValidator(field_set)
    validator.validate(field)
    assert field in validator.valid_set


def test_set_validator_empty_ok():
    validator = SetValidator(['foo'], empty_ok=True)
    validator.validate('')
    assert '' in validator.valid_set


@pytest.mark.parametrize('field_set, field', [
    ([], 'bar'),
    (['foo'], 'bar'),
    (['foo', 'bar'], 'baz'),
])
def test_set_validator_fails(field_set, field):
    validator = SetValidator(field_set)
    with pytest.raises(ValidationException):
        validator.validate(field)

    assert validator.bad == {field}


@pytest.mark.parametrize('fields, row, unique_with', [
    ([], {}, []),
    (['foo', 'bar'], {}, []),
    (['foo', 'foo'], FakeRow({'some_field': ['bar', 'baz']}), ['some_field']),
    (['foo', 'foo', 'foo'], FakeRow({
        'some_field': ['bar', 'baz', 'bar'],
        'some_other_field': ['baz', 'bar', 'bar']
    }), ['some_field', 'some_other_field']),
])
def test_unique_validator_works(fields, row, unique_with):
    validator = UniqueValidator(unique_with=unique_with)
    for field in fields:
        validator.validate(field, row)


@pytest.mark.parametrize('fields, row, unique_with, exception, bad', [
    (['foo', 'bar', 'bar'], {}, [], ValidationException, {('bar',)}),
    (['foo', 'foo'], FakeRow({'some_field': ['bar', 'bar']}), ['some_field'],
        ValidationException, {('foo', 'bar')}),
    (['foo', 'foo'], FakeRow({'some_field': ['bar', 'bar']}), ['other_field'],
        BadValidatorException, set()),
])
def test_unique_validator_fails(fields, row, unique_with, exception, bad):
    validator = UniqueValidator(unique_with=unique_with)
    with pytest.raises(exception):
        for field in fields:
            validator.validate(field, row)

    assert validator.bad == bad


@pytest.mark.parametrize('pattern, field', [
    (r'foo.*', 'foo'),
    (r'foo.*', 'foobar'),
])
def test_regex_validator_works(pattern, field):
    RegexValidator(pattern).validate(field)


def test_regex_validator_allows_empty():
    RegexValidator(r'foo.*', empty_ok=True).validate('')


@pytest.mark.parametrize('pattern, field', [
    (r'foo.*', 'afoo'),
    (r'^$', 'foo'),
])
def test_regex_validator_fails(pattern, field):
    validator = RegexValidator(pattern)
    with pytest.raises(ValidationException):
        validator.validate(field)

    assert validator.bad == {field}


def test_range_validator_works():
    RangeValidator(0, 100).validate("42")


def test_range_validator_fails():
    validator = RangeValidator(0, 100)
    with pytest.raises(ValidationException):
        validator.validate("-42")

    assert validator.bad == {'-42'}


def test_empty_validator_works():
    EmptyValidator().validate("")


def test_empty_validator_fails():
    validator = EmptyValidator()
    with pytest.raises(ValidationException):
        validator.validate("foo")

    assert validator.bad == {'foo'}


def test_non_empty_validator_works():
    NotEmptyValidator().validate("foo")


def test_non_empty_validator_fails():
    validator = NotEmptyValidator()
    with pytest.raises(ValidationException):
        validator.validate("")

    assert validator.bad == set()


def test_ignore_validator():
    validator = Ignore()
    validator.validate("foo")
    assert validator.bad is None


def test_base_class_raises():
    validator = Validator()

    with pytest.raises(NotImplementedError):
        validator.bad()

    with pytest.raises(NotImplementedError):
        validator.validate(stub(), stub())



import pytest
from pretend import stub, call, call_recorder

from ..inputs import S3File, StringIO, String, VladInput
from ..vlad import Vlad


@pytest.mark.parametrize('kwargs', [
    ({'path': 's3://some.bucket/some/s3/key.csv'}),
    ({'bucket': 'some.bucket', 'key': '/some/s3/key.csv'}),
])
def test_s3_input_works(kwargs):
    S3File(**kwargs)


@pytest.mark.parametrize('kwargs', [
    ({}),
    ({'path': 's3://some.bucket/some/s3/key.csv', 'bucket': 'some.bucket'}),
    ({'path': 's3://some.bucket/some/s3/key.csv', 'key': '/some/s3/key.csv'}),
    ({'bucket': 'some.bucket'}),
    ({'key': '/some/s3/key.csv'}),
])
def test_s3_input_fails(kwargs):
    with pytest.raises(ValueError):
        S3File(**kwargs)


@pytest.mark.parametrize('kwargs', [
    ({'string_input': 'ColA,ColB\n,'}),
    ({'string_io': StringIO('ColA,ColB\n,')}),
])
def test_string_input_works(kwargs):
    source = String(**kwargs)
    validators = {'ColA': [], 'ColB': []}
    assert Vlad(source=source, validators=validators).validate()


def test_open_s3file(monkeypatch):
    new_key = call_recorder(lambda *args, **kwargs: stub(
        get_contents_as_string=lambda: 'contents'.encode()
    ))

    get_bucket = call_recorder(lambda *args, **kwargs: stub(new_key=new_key))

    mock_boto = stub(connect_s3=lambda: stub(get_bucket=get_bucket))

    monkeypatch.setattr('vladiate.inputs.boto', mock_boto)

    result = S3File('s3://some.bucket/some/s3/key.csv').open()

    assert get_bucket.calls == [
        call('some.bucket')
    ]

    assert new_key.calls == [
        call('/some/s3/key.csv')
    ]

    assert result.readlines() == [b'contents']


def test_repr_s3file():
    s3_file = S3File('s3://some.bucket/some/s3/key.csv')
    assert repr(s3_file) == "S3File('s3://some.bucket/some/s3/key.csv')"


def test_base_class_raises():
    with pytest.raises(NotImplementedError):
        VladInput()

    class PartiallyImplemented(VladInput):
        def __init__(self):
            pass

    with pytest.raises(NotImplementedError):
        PartiallyImplemented().open()

    with pytest.raises(NotImplementedError):
        repr(PartiallyImplemented())



import pytest

from ..vlad import Vlad
from ..inputs import LocalFile, String
from ..validators import (
    EmptyValidator, FloatValidator, SetValidator, UniqueValidator
)


def test_initialize_vlad():
    source = LocalFile('vladiate/examples/vampires.csv')
    validators = {
        'Column A': [
            UniqueValidator()
        ],
        'Column B': [
            SetValidator(['Vampire', 'Not A Vampire'])
        ]
    }
    assert Vlad(source=source, validators=validators).validate()


def test_initialize_vlad_with_alternative_delimiter():
    source = LocalFile('vladiate/examples/bats.csv')
    validators = {
        'Column A': [
            UniqueValidator()
        ],
        'Column B': [
            SetValidator(['Vampire', 'Not A Vampire'])
        ]
    }
    delimiter = '|'
    vlad = Vlad(source=source, validators=validators, delimiter=delimiter)
    assert vlad.validate()


def test_initialize_vlad_no_source():
    with pytest.raises(TypeError):
        Vlad().validate()


def test_unused_validator_fails_validation():
    source = LocalFile('vladiate/examples/vampires.csv')
    validators = {
        'Column A': [
            UniqueValidator()
        ],
        'Column B': [
            SetValidator(['Vampire', 'Not A Vampire'])
        ],
        'Column C': [
            FloatValidator()
        ]
    }

    assert not Vlad(source=source, validators=validators).validate()


def test_validators_in_class_variable_are_used():
    source = LocalFile('vladiate/examples/vampires.csv')

    class TestVlad(Vlad):
        validators = {
            'Column A': [
                UniqueValidator()
            ],
            'Column B': [
                SetValidator(['Vampire', 'Not A Vampire'])
            ]
        }

    assert TestVlad(source=source).validate()


def test_missing_validators():
    source = LocalFile('vladiate/examples/vampires.csv')

    class TestVlad(Vlad):
        validators = {
            'Column A': [
                UniqueValidator()
            ],
        }

    assert not TestVlad(source=source).validate()


def test_no_fieldnames():
    source = String('')

    class TestVlad(Vlad):
        validators = {
            'Foo': [],
            'Bar': [],
        }

    assert not TestVlad(source=source).validate()


def test_fails_validation():
    source = LocalFile('vladiate/examples/vampires.csv')

    class TestVlad(Vlad):
        validators = {
            'Column A': [
                EmptyValidator()
            ],
            'Column B': [
                EmptyValidator()
            ]
        }

    vlad = TestVlad(source=source)
    assert not vlad.validate()
    assert vlad.validators['Column A'][0].fail_count == 3
    assert vlad.validators['Column B'][0].fail_count == 3


def test_gt_99_failures():
    source = String('\n'.join(['Foo'] + [str(x) for x in range(100)]))

    class TestVlad(Vlad):
        validators = {
            'Foo': [
                EmptyValidator()
            ]
        }

    assert not TestVlad(source=source).validate()


def test_ignore_missing_validators():
    source = LocalFile('vladiate/examples/vampires.csv')

    class TestVlad(Vlad):
        validators = {
            'Column A': [
                UniqueValidator()
            ],
        }

    vlad = TestVlad(source=source, ignore_missing_validators=True)

    assert vlad.validate()
    assert vlad.missing_validators == {'Column B'}



from vladiate import Vlad
from vladiate.validators import UniqueValidator, SetValidator
from vladiate.inputs import LocalFile


class YourFirstValidator(Vlad):
    source = LocalFile('vampires.csv')
    validators = {
        'Column A': [
            UniqueValidator()
        ],
        'Column B': [
            SetValidator(['Vampire', 'Not A Vampire'])
        ]
    }


class YourFirstNonCommaDelimitedValidator(Vlad):
    source = LocalFile('bats.csv')
    validators = {
        'Column A': [
            UniqueValidator()
        ],
        'Column B': [
            SetValidator(['Vampire', 'Not A Vampire'])
        ]
    }
    delimiter = '|'


class YourFirstFailingValidator(Vlad):
    source = LocalFile('potential_vampires.csv')
    validators = {
        'Column A': [
            UniqueValidator()
        ],
        'Column B': [
            SetValidator(['Vampire', 'Not A Vampire'])
        ]
    }


class YourFirstEmptyValidator(Vlad):
    source = LocalFile('real_vampires.csv')
    validators = {}


class YourSecondEmptyValidator(Vlad):
    source = LocalFile('real_vampires.csv')
    validators = {
        'Column A': [],
        'Column B': [],
        'Column C': [],
    }



import os
import sys
import inspect

import pytest
from pretend import stub, call, call_recorder

from ..main import (  # NOQA
    parse_args, is_vlad, find_vladfile, load_vladfile, _vladiate, main, run,
    _is_package
)

from ..vlad import Vlad
from ..inputs import String
from ..examples import vladfile
from ..examples.vladfile import YourFirstFailingValidator


def test_parse_args():
    options = parse_args()

    assert options.list_commands is False
    assert options.processes == 1
    assert options.show_version is False
    assert options.vladfile == 'vladfile'
    assert options.vlads == ['test']


@pytest.mark.parametrize('tup, expected', [
    (('foo', 'bar'), False),
    (('YourFirstFailingValidator', YourFirstFailingValidator), True),
])
def test_is_vlad(tup, expected):
    assert is_vlad(tup) == expected


@pytest.mark.parametrize('name, path', [
    ('vladfile', './vladiate/examples/'),
    ('./vladiate/examples/vladfile', '.'),
])
def test_find_vladfile(name, path):
    assert find_vladfile(name, path)


@pytest.mark.parametrize('path', [
    ('./vladiate/examples/vladfile'),
    (os.path.join(sys.path[0], 'vladfile')),
    (os.path.join(sys.path[1], 'vladfile')),
])
def test_load_vladfile(path):
    doc, vlads = load_vladfile(path)
    available = dict(inspect.getmembers(
        vladfile,
        lambda x: inspect.isclass(x) and issubclass(x, Vlad) and x != Vlad
    ))

    assert set(vlads.keys()) == set(available.keys())


def test_vladiate(monkeypatch):
    put = call_recorder(lambda *args, **kwargs: stub())
    result_queue = stub(put=put)

    validate_result = stub()

    monkeypatch.setattr('vladiate.main.result_queue', result_queue)

    class TestVlad(Vlad):
        source = String('foo')
        validators = {}

        def validate(self):
            return validate_result

    _vladiate(TestVlad)

    assert put.calls == [
        call(validate_result)
    ]


def test_run(monkeypatch):
    main_ret = stub()
    main = call_recorder(lambda: main_ret)
    exit = call_recorder(lambda x: stub())
    monkeypatch.setattr('vladiate.main.main', main)
    try:
        monkeypatch.setattr('__builtin__.exit', exit)
    except:
        monkeypatch.setattr('builtins.exit', exit)

    run('__main__')

    assert main.calls == [
        call()
    ]

    assert exit.calls == [
        call(main_ret)
    ]


def test_main_when_get_nowait_raises(monkeypatch):
    monkeypatch.setattr(
        'vladiate.main.parse_args', lambda: stub(
            list_commands=False,
            show_version=False,
            vladfile=stub(),
            vlads=['Something'],
            processes=2,
        )
    )
    monkeypatch.setattr(
        'vladiate.main.find_vladfile', lambda *args, **kwargs: stub()
    )
    vlad = call_recorder(lambda *args, **kwargs: stub(validate=lambda: stub()))
    vlad.source = stub()

    monkeypatch.setattr(
        'vladiate.main.load_vladfile',
        lambda *args, **kwargs: (None, {'Something': vlad})
    )

    Pool = call_recorder(
        lambda *args, **kwargs: stub(map=lambda *args, **kwargs: stub())
    )
    monkeypatch.setattr('vladiate.main.Pool', Pool)

    class MockEmpty(BaseException):
        pass

    def _raise_empty():
        raise MockEmpty

    result_queue = stub(get_nowait=_raise_empty)
    monkeypatch.setattr('vladiate.main.result_queue', result_queue)
    monkeypatch.setattr('vladiate.main.Empty', MockEmpty)

    assert main() is os.EX_OK


@pytest.mark.parametrize('get_nowait, expected', [
    (lambda: stub(), os.EX_OK),
    (lambda: False, os.EX_DATAERR),
])
def test_main_with_multiprocess(monkeypatch, get_nowait, expected):
    monkeypatch.setattr(
        'vladiate.main.parse_args', lambda: stub(
            list_commands=False,
            show_version=False,
            vladfile=stub(),
            vlads=['Something'],
            processes=2,
        )
    )
    monkeypatch.setattr(
        'vladiate.main.find_vladfile', lambda *args, **kwargs: stub()
    )
    vlad = call_recorder(lambda *args, **kwargs: stub(validate=lambda: stub()))
    vlad.source = stub()

    monkeypatch.setattr(
        'vladiate.main.load_vladfile',
        lambda *args, **kwargs: (None, {'Something': vlad})
    )

    Pool = call_recorder(
        lambda *args, **kwargs: stub(map=lambda *args, **kwargs: stub())
    )
    monkeypatch.setattr('vladiate.main.Pool', Pool)
    result_queue = stub(get_nowait=get_nowait)
    monkeypatch.setattr('vladiate.main.result_queue', result_queue)

    assert main() is expected


def test_main_with_vlads_in_args(monkeypatch):
    monkeypatch.setattr(
        'vladiate.main.parse_args', lambda: stub(
            list_commands=False,
            show_version=False,
            vladfile=stub(),
            vlads=['Something'],
            processes=1,
        )
    )
    monkeypatch.setattr(
        'vladiate.main.find_vladfile', lambda *args, **kwargs: stub()
    )
    vlad = call_recorder(lambda *args, **kwargs: stub(validate=lambda: stub()))
    vlad.source = stub()

    monkeypatch.setattr(
        'vladiate.main.load_vladfile',
        lambda *args, **kwargs: (None, {'Something': vlad})
    )
    assert main() is None


def test_main_no_vlads_in_args(monkeypatch):
    monkeypatch.setattr(
        'vladiate.main.parse_args', lambda: stub(
            list_commands=False,
            show_version=False,
            vladfile=stub(),
            vlads=[],
            processes=1,
        )
    )
    monkeypatch.setattr(
        'vladiate.main.find_vladfile', lambda *args, **kwargs: stub()
    )

    vlad = call_recorder(lambda *args, **kwargs: stub(validate=lambda: stub()))
    vlad.source = stub()
    monkeypatch.setattr(
        'vladiate.main.load_vladfile',
        lambda *args, **kwargs: (None, {'Something Else': vlad})
    )
    assert main() is None


def test_main_missing_vlads(monkeypatch):
    monkeypatch.setattr(
        'vladiate.main.parse_args', lambda: stub(
            list_commands=False,
            show_version=False,
            vladfile=stub(),
            vlads=['Something'],
        )
    )
    monkeypatch.setattr(
        'vladiate.main.find_vladfile', lambda *args, **kwargs: stub()
    )
    monkeypatch.setattr(
        'vladiate.main.load_vladfile',
        lambda *args, **kwargs: (None, {'Something Else': stub()})
    )
    assert main() == os.EX_UNAVAILABLE


def test_main_no_vlads_loaded(monkeypatch):
    monkeypatch.setattr(
        'vladiate.main.parse_args', lambda: stub(
            list_commands=False,
            show_version=False,
            vladfile=stub(),
        )
    )
    monkeypatch.setattr(
        'vladiate.main.find_vladfile', lambda *args, **kwargs: stub()
    )
    vlads = []
    monkeypatch.setattr(
        'vladiate.main.load_vladfile', lambda *args, **kwargs: (None, vlads)
    )
    assert main() == os.EX_NOINPUT


def test_main_list_commands(monkeypatch):
    monkeypatch.setattr(
        'vladiate.main.parse_args', lambda: stub(
            list_commands=True,
            show_version=False,
            vladfile=stub(),
        )
    )
    monkeypatch.setattr(
        'vladiate.main.find_vladfile', lambda *args, **kwargs: stub()
    )
    vlads = ['Something']
    monkeypatch.setattr(
        'vladiate.main.load_vladfile', lambda *args, **kwargs: (None, vlads)
    )
    assert main() == os.EX_OK


def test_main_show_version(monkeypatch):
    monkeypatch.setattr(
        'vladiate.main.parse_args', lambda: stub(show_version=True)
    )
    assert main() == os.EX_OK


def test_main_no_vladfile(monkeypatch):
    monkeypatch.setattr(
        'vladiate.main.find_vladfile', lambda *args, **kwargs: None
    )
    assert main() == os.EX_NOINPUT


@pytest.mark.parametrize('path, expected', [
    ('foo/bar', False),
    ('vladiate/test', True),
])
def test_is_package(path, expected):
    assert _is_package(path) == expected
